/*
 *	field.c
 *
 *	* Sun Jan 20 18:22:00 JST 2002 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

static void cell_draw(CELL* cell, int x, int y);
static void ctrl_draw(char color[2][2], int x, int y, int rot);
static void field_next(FIELD* field);
static int flip_check(CELL cells[FIELD_ROWS][FIELD_COLS], int x, int y, FLIP flip[10]);
static int flip_check_1(CELL cells[FIELD_ROWS][FIELD_COLS], int x, int y, int vx, int vy);
static void flip_seek(FIELD* field);

/* ]ɂ𐶐邩H
 * {͔]ɂ𐶐܂AP/ECEł͌Â炢łB
 * ̃V{LɂƁA{Ɠ𐶐܂B
 */
/*#define FLIP_STAR*/

/****************************************************************************
 *	[Jϐ
 ****************************************************************************/

/* color -> L */
typedef struct _PAT {
	int x, y, w, h;
} PAT;
static PAT soldam_pat[] = {
	{ 0, 0, 0, 0 },
	{ CHR(SOLDAM0) },
	{ CHR(SOLDAM1) },
	{ CHR(SOLDAM2) },
	{ CHR(SOLDAM3) },
};
static PAT erase_pat[] = {
	{ CHR(ERASE0) },
	{ CHR(ERASE1) },
	{ CHR(ERASE2) },
	{ CHR(ERASE3) },
};
static PAT light_pat[] = {
	{ CHR(LIGHT0) },
	{ CHR(LIGHT1) },
	{ CHR(LIGHT2) },
	{ CHR(LIGHT1) },
};
static PAT star_pat[] = {
	{ CHR(STAR00) },
	{ CHR(STAR10) },
	{ CHR(STAR01) },
	{ CHR(STAR11) },
	{ CHR(STAR02) },
	{ CHR(STAR12) },
};

/* 2 COLORS MODE: +
 * 3 COLORS MODE: ++
 * 4 COLORS MODE: +++
 */
static next_gen[4] = { 1, 4, 3, 2 };

/* ]`FbN̎n_ItZbg(0,1)ƕ(2,3) */
static int flip_check_tbl[10][4] = {
	{ 0, 0, -1, -1 },
	{ 0, 0, -1,  0 },
	{ 0, 1, -1,  0 },
	{ 0, 1, -1,  1 },
	{ 0, 1,  0,  1 },
	{ 1, 1,  0,  1 },
	{ 1, 1,  1,  1 },
	{ 1, 1,  1,  0 },
	{ 1, 0,  1,  0 },
	{ 1, 0,  1, -1 }
};

/* x -> Xs[h */
static int speed_tbl[100] = {
	 16,  16,  32,  32,  64,  64,  96,  96, 128, 128,	/*  0` 9 */
	 32,  32,  64,  64,  96,  96, 128, 128, 160, 160,	/* 10`19 */
	 32,  32,  64,  64,  96,  96, 128, 128, 160, 160,	/* 20`29 */
	 64,  64,  96,  96, 128, 128, 160, 160, 192, 192,	/* 30`39 */
	 64,  64,  96,  96, 128, 128, 160, 160, 192, 192,	/* 40`49 */
	 96,  96, 128, 128, 160, 160, 192, 192, 224, 224,	/* 50`59 */
	 96,  96, 128, 128, 160, 160, 192, 192, 224, 224,	/* 60`69 */
	128, 128, 160, 160, 192, 192, 224, 224, 256, 256,	/* 70`79 */
	128, 128, 160, 160, 192, 192, 224, 224, 256, 256,	/* 80`89 */
	160, 160, 192, 192, 224, 224, 256, 256, 256, 256,	/* 90`99 */
};

/* 1,2,...,FIELD_ROWSs̊es̃XRA */
static int score_tbl[FIELD_ROWS] = {
	  10,   20,   40,   60,   80,
	 100,  200,  400,  600,  800,
	1000, 2000, 4000, 6000, 8000,
};
#define SCORE_CTRL	1	/* ɖXRA */

/*  */
#define STARS_MAX	32
typedef struct _STAR {
	int use;
	int x, y;
	int count;
} STAR;
static STAR stars[STARS_MAX];
static void star_new(int x, int y);
static void star_draw(STAR* star);

/* ΂΂Z */
#define BARAS_MAX	(FIELD_COLS * FIELD_ROWS)
#define BARA_G		0x20	/* d͉x(8.8rbg) */
#define BARA_R		0.5	/* ł̔˗ */
typedef struct _BARA {
	char color;	/* F(0:I[) */
	int x, y;	/* ʒu(8.8rbg) */
	int vx, vy;	/* x(8.8rbg) */
	int bounce;	/* ˂ */
} BARA;
static BARA baras[BARAS_MAX];

/****************************************************************************
 *	field_draw
 ****************************************************************************/

void
field_draw(FIELD* field)
{
	int col, row, x, y, i, j;
	CELL* cell;
	STAR* star;
	BARA* bara;
	PAT* pat;
	DRAW_OBJECT obj;

	/* wi */
	pceLCDSetObject(&obj, &bmp[1],
		0,                                      0,
		CHR_BACKGROUND_X,                       CHR_BACKGROUND_Y,                     FIELD_X,            CHR_BACKGROUND_H,
		DRW_NOMAL);
	pceLCDDrawObject(obj);
	pceLCDSetObject(&obj, &bmp[1],
		CHR_BACKGROUND_X + (FIELD_X + FIELD_W), 0,
		CHR_BACKGROUND_X + (FIELD_X + FIELD_W), CHR_BACKGROUND_Y, CHR_BACKGROUND_W - (FIELD_X + FIELD_W), CHR_BACKGROUND_H,
		DRW_NOMAL);
	pceLCDDrawObject(obj);

	/* ubN */
	switch(field->block_offset & 3) {
	case 0:
		y = FIELD_Y;
		i = 0;
		break;
	case 1:
		y = FIELD_Y - CELL_UNIT;
		i = 1;
		break;
	case 2:
		y = FIELD_Y;
		i = 1;
		break;
	default:
		y = FIELD_Y - CELL_UNIT;
		i = 0;
		break;
	}
	while(y < CLEARSOLDAM_Y) {
		x = FIELD_X;
		j = i;
		for(col = 0; col < FIELD_COLS; col += 2) {
			if(j == 0) {
				pceLCDSetObject(&obj, &bmp[1], x, y, CHR(BLOCK0), DRW_NOMAL);
			} else {
				pceLCDSetObject(&obj, &bmp[1], x, y, CHR(BLOCK1), DRW_NOMAL);
			}
			pceLCDDrawObject(obj);
			x += BLOCK_UNIT;
			j ^= 1;
		}
		y += BLOCK_UNIT;
		i ^= 1;
	}

	/* NA\_ */
	pceLCDSetObject(&obj, &bmp[1], CLEARSOLDAM_X, CLEARSOLDAM_Y, CHR(CLEARSOLDAM), DRW_NOMAL);
	pceLCDDrawObject(obj);

	/* \_ */
	switch(field->mode) {
	case FM_OVER_INIT:
	case FM_OVER:
		bara = baras;
		for(i = 0; i < BARAS_MAX; i++) {
			if(bara->color == 0) break; /* ȏȂ */
			pat = &soldam_pat[(int)bara->color];
			pceLCDSetObject(&obj, &bmp[0], bara->x >> 8, bara->y >> 8, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
			bara++;
		}
		break;
	default:
		for(i = 0; i < 2; i++) {
			cell = &field->cells[0][0];
			y = FIELD_Y;
			for(row = 0; row < FIELD_ROWS; row++) {
				x = FIELD_X;
				for(col = 0; col < FIELD_COLS; col++) {
					if(cell->color != 0) {
						if(i == 0) {
							if(cell->mode != CM_CLEAR) cell_draw(cell, x, y);
						} else {
							if(cell->mode == CM_CLEAR) cell_draw(cell, x, y);
						}
					}
					cell++;
					x += CELL_UNIT;
				}
				y += CELL_UNIT;
			}
		}
		break;
	}

	/* Z */
	switch(field->mode) {
	case FM_CTRL_FLOAT:
	case FM_CTRL:
	case FM_CTRL_LAND:
		x = FIELD_X + CELL_UNIT * field->x + field->move_x;
		y = FIELD_Y + CELL_UNIT * field->y + field->move_y;
		ctrl_draw(field->ctrl, x, y, field->rot);
		pat = &light_pat[timer & 3]; /* 䒆͒SɌ` */
		pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		break;
	}

	/*  */
	star = stars;
	for(i = 0; i < STARS_MAX; i++) {
		if(star->use) star_draw(star);
		star++;
	}

	/* ~bg */
	if(timer & 1) pceLCDPaint(3, LIMIT_X, LIMIT_Y, LIMIT_W, LIMIT_H);

	/* NEXTZ */
	pceLCDSetObject(&obj, &bmp[1], NEXT_X, NEXT_Y, CHR(NEXT), DRW_NOMAL);
	pceLCDDrawObject(obj);
	ctrl_draw(field->next, NEXT_X + 3, NEXT_Y + 3, 0);

	pceFontSetType(0x82);
	pceFontSetTxColor(0);
	pceFontSetBkColor(-1);

	/* HI-SCORE */
	pceLCDPaint(3, HISCORE_X, HISCORE_Y, HISCORE_W, HISCORE_H);
	pceFontSetPos(HISCORE_X + 1, HISCORE_Y + 1);
	pceFontPutStr("HI");
	pceFontSetPos(HISCORE_X + 10, HISCORE_Y + 1);
	pceFontPrintf("%6d", hi_score > field->score ? hi_score : field->score);

	/* 1P-SCORE */
	pceLCDPaint(3, P1SCORE_X, P1SCORE_Y, P1SCORE_W, P1SCORE_H);
	pceFontSetPos(P1SCORE_X + 1, P1SCORE_Y + 1);
	pceFontPutStr("1P");
	pceFontSetPos(P1SCORE_X + 10, P1SCORE_Y + 1);
	pceFontPrintf("%6d", field->score);

	/* 1P-SCORE */
	pceLCDPaint(3, LINE_X, LINE_Y, LINE_W, LINE_H);
	pceFontSetPos(LINE_X + 1, LINE_Y + 1);
	pceFontPutStr("LINE");
	pceFontSetPos(LINE_X + 18, LINE_Y + 1);
	pceFontPrintf("%4d", field->line);

	/* x */
	pceFontSetType(0x80);
	pceLCDPaint(3, LEVEL_X, LEVEL_Y, LEVEL_W, LEVEL_H);
	pceFontSetPos(LEVEL_X + (LEVEL_W - 5 * 5) / 2, LEVEL_Y + 0);
	pceFontPutStr("LEVEL");
	pceFontSetType(0x81);
	if(field->level < 10) {
		pceFontSetPos(LEVEL_X + (LEVEL_W - 8 * 1) / 2, LEVEL_Y + 8);
	} else if(field->level < 100) {
		pceFontSetPos(LEVEL_X + (LEVEL_W - 8 * 2) / 2, LEVEL_Y + 8);
	} else if(field->level < 1000) {
		pceFontSetPos(LEVEL_X + (LEVEL_W - 8 * 3) / 2, LEVEL_Y + 8);
	} else {
		pceFontSetPos(LEVEL_X + (LEVEL_W - 8 * 4) / 2, LEVEL_Y + 8);
	}
	pceFontPrintf("%d", field->level);

	/* J[[h */
	switch(field->mode) {
	case FM_COLORMODE:
		pceFontSetType(0x82);
		pceFontSetTxColor(0);
		pceFontSetBkColor(-1);
		if(timer & 1) {
			pceLCDPaint(3, COLORMODE_X, COLORMODE_Y, COLORMODE_W, COLORMODE_H);
		} else {
			pceLCDPaint(2, COLORMODE_X, COLORMODE_Y, COLORMODE_W, COLORMODE_H);
		}
		pceFontSetPos(COLORMODE_X + 1, COLORMODE_Y + 1);
		pceFontPrintf("%d COLORS MODE", field->colors);
		break;
	}
}

/****************************************************************************
 *	cell_draw
 ****************************************************************************/

static void
cell_draw(CELL* cell, int x, int y)
{
	int n;
	PAT* pat;
	DRAW_OBJECT obj;

	switch(cell->mode) {
	case CM_NORMAL:
		pat = &soldam_pat[(int)cell->color];
		pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		break;
	case CM_FLIP0:
	case CM_FLIP1:
	case CM_FLIP2:
	case CM_FLIP3:
		pat = &soldam_pat[(int)cell->mode]; /* CM_FLIPx̒l͕ωcolorlƓ */
		n = cell->count;
		if(n < 0) {
			n = 0;
		} else if(n > pat->w) {
			n = pat->w;
		}
		if(n > 0) {
			pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, n, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
		}
		if(n < pat->w) {
			pat = &soldam_pat[(int)cell->color];
			pceLCDSetObject(&obj, &bmp[0], x + n, y, pat->x + n, pat->y, pat->w - n, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
		}
		break;
	case CM_ERASE:
		n = cell->count;
		if(n < 0) {
			pat = &soldam_pat[(int)cell->color];
			pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
		} else if(n < 4) {
			pat = &erase_pat[n];
			pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
		}
		break;
	case CM_CLEAR:
		pat = &soldam_pat[(int)cell->color];
		n = cell->count;
		if(n < 0) {
			pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
		} else {
			y += n * CELL_UNIT;
			if(y > CLEARSOLDAM_Y) y = CLEARSOLDAM_Y;
			pceLCDSetObject(&obj, &bmp[0], x, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
			pceLCDDrawObject(obj);
		}
		break;
	}
}

/****************************************************************************
 *	ctrl_draw
 ****************************************************************************/

static void
ctrl_draw(char color[2][2], int x, int y, int rot)
{
	PAT* pat;
	DRAW_OBJECT obj;

	if(rot >= 0) {
		pat = &soldam_pat[(int)color[0][0]];
		pceLCDSetObject(&obj, &bmp[0], x + rot, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		pat = &soldam_pat[(int)color[0][1]];
		pceLCDSetObject(&obj, &bmp[0], x + CELL_UNIT, y + rot, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		pat = &soldam_pat[(int)color[1][1]];
		pceLCDSetObject(&obj, &bmp[0], x + CELL_UNIT - rot, y + CELL_UNIT, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		pat = &soldam_pat[(int)color[1][0]];
		pceLCDSetObject(&obj, &bmp[0], x, y + CELL_UNIT - rot, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
	} else {
		pat = &soldam_pat[(int)color[0][0]];
		pceLCDSetObject(&obj, &bmp[0], x, y - rot, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		pat = &soldam_pat[(int)color[0][1]];
		pceLCDSetObject(&obj, &bmp[0], x + CELL_UNIT + rot, y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		pat = &soldam_pat[(int)color[1][1]];
		pceLCDSetObject(&obj, &bmp[0], x + CELL_UNIT, y + CELL_UNIT + rot, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
		pat = &soldam_pat[(int)color[1][0]];
		pceLCDSetObject(&obj, &bmp[0], x - rot, y + CELL_UNIT, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
		pceLCDDrawObject(obj);
	}
}

/****************************************************************************
 *	field_next
 ****************************************************************************/

void
field_next(FIELD* field)
{
	int col, row, n;

	/* NEXTZ𐧌Zֈڂ܂B */
	memcpy(field->ctrl, field->next, 2 * 2);

	/* VNEXTZ߂܂B */
	for(row = 0; row < 2; row++) {
		for(col = 0; col < 2; col++) {
			n = RAND(field->seed) % field->colors;
			field->next[row][col] = next_gen[n];
		}
	}
}

/****************************************************************************
 *	field_init
 ****************************************************************************/

void
field_init(FIELD* field)
{
	memset(stars, 0, sizeof stars);

	memset(field, 0, sizeof(FIELD));
	field->level = 1;
	field->colors = 2; /* field_next()ɂȂƁA[ZG[ɂȂ܂ */
	field->speed = CELL_UNIT * speed_tbl[field->level];
	field->seed = pceTimerGetCount();
	field_next(field); /* NEXT */
}

/****************************************************************************
 *	field_proc
 ****************************************************************************/

int
field_proc(FIELD* field)
{
	int b, i, j, x, y;
	char color;
	CELL (*cells)[FIELD_COLS], *cell;
	FLIP* flip;
	BARA* bara;

	cells = field->cells;

	switch(field->mode) {
	case FM_TURN_INIT:
		b = 0; /* J[[hύXꂽ1ɂȂ */
		while(field->n_erase != 0) {
			/* CvCɈڂB */
			field->n_erase--;
			field->line++;

			/* 3C1xAbvB */
			if(field->line % 3 == 0) {
				field->level++;

				/* Xs[h̕ύXB */
				if(field->level < 100) {
					field->speed = CELL_UNIT * speed_tbl[field->level];
				} else {
					field->speed = CELL_UNIT * speed_tbl[100 - 1];
				}

				/* J[[h̕ύXB */
				switch(field->level) {
				case LEVEL_3COLORS:
					field->colors = 3;
					b = 1; /* J[[hύX */
					break;
				case LEVEL_4COLORS:
					field->colors = 4;
					b = 1; /* J[[hύX */
					break;
				}
			}
		}
		if(b) { /* J[[hύX */
			field->mode = FM_COLORMODE;
			field->count = 0;
		} else {
			field->mode = FM_CTRL_INIT;
			field->count = 0;
		}
		break;

	case FM_COLORMODE:
		field->count++;
		if(field->count >= COLORMODE_TIME) {
			field->mode = FM_CTRL_INIT;
			field->count = 0;
		}
		break;

	case FM_CTRL_INIT:
		/* Zo\ׂ܂B */
		field->x = FIELD_COLS / 2 & ~1;
		field->y = 0;
		if(cells[field->y + 0][field->x + 0].color != 0 || cells[field->y + 0][field->x + 1].color != 0 ||
		   cells[field->y + 1][field->x + 0].color != 0 || cells[field->y + 1][field->x + 1].color != 0) {
			field->mode = FM_OVER_INIT;
			field->count = 0;
			return 0;
		}

		/* Z擾܂B */
		field_next(field);

		/* ]XV܂B */
		flip_seek(field);
		field->star = 0;

		/* (V)ֈڍs܂B */
		field->mode = FM_CTRL_FLOAT;
		field->count = 0;
		field->rot = 0;
		field->move_x = 0;
		field->move_y = 0;
		break;

	case FM_CTRL_FLOAT:
	case FM_CTRL:
	case FM_CTRL_LAND:
		/* B */
		flip = field->flip;
		for(i = 0; i < field->n_flip; i++) {
			if(field->star <= flip->n + 1) {
				x = flip->x + flip->vx * field->star;
				y = flip->y + flip->vy * field->star;
				star_new(FIELD_X + CELL_UNIT * x, FIELD_Y + CELL_UNIT * y);
			}
			flip++;
		}
		field->star++;
		if(field->star >= STAR_INTERVAL) field->star = 0;

		/* ] */
		if(field->rot >= 0) {
			if(pad & TRG_A) {
				color = field->ctrl[0][0];
				field->ctrl[0][0] = field->ctrl[1][0];
				field->ctrl[1][0] = field->ctrl[1][1];
				field->ctrl[1][1] = field->ctrl[0][1];
				field->ctrl[0][1] = color;
				field->rot -= CELL_UNIT;
				flip_seek(field); /* ]XV܂B */
			}
		}
		if(field->rot <= 0) {
			if(pad & TRG_B) {
				color = field->ctrl[0][0];
				field->ctrl[0][0] = field->ctrl[0][1];
				field->ctrl[0][1] = field->ctrl[1][1];
				field->ctrl[1][1] = field->ctrl[1][0];
				field->ctrl[1][0] = color;
				field->rot += CELL_UNIT;
				flip_seek(field); /* ]XV܂B */
			}
		}
		if(field->rot < 0) {
			if((field->rot += ROT_SPEED) > 0) field->rot = 0;
		} else if(field->rot > 0) {
			if((field->rot -= ROT_SPEED) < 0) field->rot = 0;
		}

		/* Eړ */
		if(field->move_x <= 0) {
			if((field->x >= 2) && (pad & PAD_LF)) {
				if(cells[field->y + 0][field->x - 2 + 0].color == 0 && cells[field->y + 0][field->x - 2 + 1].color == 0 &&
				   cells[field->y + 1][field->x - 2 + 0].color == 0 && cells[field->y + 1][field->x - 2 + 1].color == 0) {
					field->x -= 2;
					field->move_x += CELL_UNIT * 2;
					flip_seek(field); /* ]XV܂B */
				}
			}
		}
		if(field->move_x >= 0) {
			if((field->x <= FIELD_COLS - 4) && (pad & PAD_RI)) {
				if(cells[field->y + 0][field->x + 2 + 0].color == 0 && cells[field->y + 0][field->x + 2 + 1].color == 0 &&
				   cells[field->y + 1][field->x + 2 + 0].color == 0 && cells[field->y + 1][field->x + 2 + 1].color == 0) {
					field->x += 2;
					field->move_x -= CELL_UNIT * 2;
					flip_seek(field); /* ]XV܂B */
				}
			}
		}
		if(field->move_x < 0) {
			if((field->move_x += MOVE_SPEED) > 0) field->move_x = 0;
		} else if(field->move_x > 0) {
			if((field->move_x -= MOVE_SPEED) < 0) field->move_x = 0;
		}

		switch(field->mode) {
		case FM_CTRL_FLOAT:
			/* Jn`FbNB */
			field->count++;
			if((field->count >= FLOAT_TIME) || (pad & PAD_DN)) {
				field->mode = FM_CTRL;
				field->count = 0;
			}
			break;
		case FM_CTRL:
			/* B */
			field->count += field->speed;
			if(field->count >= 0x100) {
				field->move_y += field->count >> 8;
				field->count &= 0xFF;
			}
			if(pad & PAD_DN) {
				field->move_y = 0;
				field->count = 0;
				field->score += SCORE_CTRL;
			}
			/* n`FbNB */
			while(field->move_y >= 0) {
				if(field->y == FIELD_ROWS - 3 ||
				   cells[field->y + 2][field->x + 0].color != 0 || cells[field->y + 2][field->x + 1].color != 0) {
					field->mode = FM_CTRL_LAND;
					field->count = 0;
					field->move_y = 0;
					break; /* ݒuAȏ㗎Ȃ */
				} else {
					field->y++;
					field->move_y -= CELL_UNIT;
				}
			}
			break;
		case FM_CTRL_LAND:
			/* ڒ`FbNɗĊJ`FbNȂƁA󒆂Ōł܂邱Ƃ܂B */
			if(field->y < FIELD_ROWS - 3 &&
			   cells[field->y + 2][field->x + 0].color == 0 && cells[field->y + 2][field->x + 1].color == 0) {
				field->mode = FM_CTRL;
				field->count = 0;
				break; /* ĊJAڒ`FbN͍sȂ */
			}
			/* ꂩAڒ`FbNB */
			field->count++;
			if(field->count >= LAND_TIME) {
				cells[field->y + 0][field->x + 0].color = field->ctrl[0][0];
				cells[field->y + 0][field->x + 0].mode = CM_NORMAL;
				cells[field->y + 0][field->x + 1].color = field->ctrl[0][1];
				cells[field->y + 0][field->x + 1].mode = CM_NORMAL;
				cells[field->y + 1][field->x + 0].color = field->ctrl[1][0];
				cells[field->y + 1][field->x + 0].mode = CM_NORMAL;
				cells[field->y + 1][field->x + 1].color = field->ctrl[1][1];
				cells[field->y + 1][field->x + 1].mode = CM_NORMAL;
				field->mode = FM_FLIP_INIT;
				field->count = 0;
			}
			break;
		}
		break;

	case FM_FLIP_INIT:
		/* ]`FbNB */
		field->n_flip = flip_check(cells, field->x, field->y, field->flip);
		if(field->n_flip != 0) {
			b = 0; /* NA\_g1ɂȂ */
			flip = field->flip;
			for(i = 0; i < field->n_flip; i++) {
				x = flip->x;
				y = flip->y;
				color = cells[y][x].color;
				for(j = 0; j < flip->n; j++) {
					x += flip->vx;
					y += flip->vy;
					cells[y][x].mode = color; /* CM_FLIPx = color */
					cells[y][x].count = -j; /* قǒxĔ] */
				}
				if(flip->clear) b = 1; /* NA\_gp */
				flip++;
			}
			if(b) { /* NA\_ */
				y = FIELD_ROWS - 1;
				for(x = 0; x < FIELD_COLS; x++) {
					cells[y][x].mode = CM_ERASE;
					cells[y][x].count = 0;
				}
			}
			field->mode = FM_FLIP;
			field->count = 0;
#ifdef FLIP_STAR
			field->star = 0; /* FM_FLIPł̐̂ */
#endif /*FLIP_STAR*/
		} else {
			field->mode = FM_ERASE_INIT;
			field->count = 0;
		}
		break;

	case FM_FLIP:
#ifdef FLIP_STAR
		/* B */
		flip = field->flip;
		for(i = 0; i < field->n_flip; i++) {
			if(field->star <= flip->n + 1) {
				x = flip->x + flip->vx * field->star;
				y = flip->y + flip->vy * field->star;
				star_new(FIELD_X + CELL_UNIT * x, FIELD_Y + CELL_UNIT * y);
			}
			flip++;
		}
		field->star++;
		/* ]tF[Ył͈͐񂫂łB */
#endif /*FLIP_STAR*/

		b = 0; /* ł쒆Ȃ1ɂȂ */
		cell = &cells[0][0];
		for(y = 0; y < FIELD_ROWS; y++) {
			for(x = 0; x < FIELD_COLS; x++) {
				switch(cell->mode) {
				case CM_FLIP0:
				case CM_FLIP1:
				case CM_FLIP2:
				case CM_FLIP3:
					b = 1; /* 쒆 */
					cell->count++;
					if(cell->count >= CELL_UNIT) {
						cell->color = cell->mode; /* color = CM_FLIPx */
						cell->mode = CM_NORMAL;
					}
					break;
				case CM_ERASE: /* NA\_̂ */
					b = 1; /* 쒆 */
					cell->count++;
					if(cell->count >= CELL_UNIT) { /* TODO: */
						cell->color = 0;
						cell->mode = CM_NORMAL;
					}
					break;
				}
				cell++;
			}
		}
		if(!b) { /* 쒆̃ZȂȂ... */
			field->mode = FM_ERASE_INIT;
			field->count = 0;
		}
		break;

	case FM_ERASE_INIT:
		/*field->n_erase = 0;ɂȂĂnY*/
		for(y = 0; y < FIELD_ROWS - 1; y++) { /* NA\_ */
			color = cells[y][0].color;
			if(color != 0) {
				for(x = 1; x < FIELD_COLS; x++) {
					if(cells[y][x].color != color) break;
				}
				if(x == FIELD_COLS) {
					if(field->n_erase == 0) {
						/* ԏ̏ŃCNA\_ɂȂ */
						for(x = 0; x < FIELD_COLS; x++) {
							cells[y][x].mode = CM_CLEAR;
							cells[y][x].count = 0;
						}
					} else {
						/* ȊO͒PɏłB */
						for(x = 0; x < FIELD_COLS; x++) {
							cells[y][x].mode = CM_ERASE;
							cells[y][x].count = 0;
						}
					}
					field->score += score_tbl[field->n_erase];
					field->n_erase++;
				}
			}
		}
		if(field->n_erase != 0) { /* 1Cłł... */
			/* ̃NA\_mɏłB */
			y = FIELD_ROWS - 1;
			for(x = 0; x < FIELD_COLS; x++) {
				if(cells[y][x].color != 0) cells[y][x].mode = CM_ERASE;
			}
			field->mode = FM_ERASE;
			field->count = 0;
		} else {
			field->mode = FM_FALL_INIT;
			field->count = 0;
		}
		break;

	case FM_ERASE:
		b = 0; /* ł쒆Ȃ1ɂȂ */
		cell = &cells[0][0];
		for(y = 0; y < FIELD_ROWS; y++) {
			for(x = 0; x < FIELD_COLS; x++) {
				switch(cell->mode) {
				case CM_ERASE:
					b = 1; /* 쒆 */
					cell->count++;
					if(cell->count >= CELL_UNIT) { /* TODO: */
						cell->color = 0;
						cell->mode = CM_NORMAL;
					}
					break;
				case CM_CLEAR:
					b = 1; /* 쒆 */
					cell->count++;
					if(y + cell->count >= FIELD_ROWS - 1) {
						cells[FIELD_ROWS - 1][x].color = cell->color;
						cells[FIELD_ROWS - 1][x].mode = CM_NORMAL;
						cell->color = 0;
						cell->mode = CM_NORMAL;
					}
					break;
				}
				cell++;
			}
		}
		if(!b) {
			field->mode = FM_FALL_INIT;
			field->count = 0;
		}
		break;

	case FM_FALL_INIT:
		/* 邱ƂȂ */
		field->mode = FM_FALL;
		field->count = 0;
		break;

	case FM_FALL:
		b = 0; /* ł1ɂȂ */
		for(x = 0; x < FIELD_COLS; x++) {
			for(y = FIELD_ROWS - 2; y >= 1; y--) { /* NA\_̈ォŏï܂ */
				if(cells[y][x].color == 0) {
					for(i = y; i >= 1; i--) {
						color = cells[i - 1][x].color;
						cells[i][x].color = color;
						if(color != 0) b = 1;
					}
					cells[i][x].color = 0;
				}
			}
		}
		if(b) {
			field->block_offset++;
		} else {
			field->mode = FM_TURN_INIT;
			field->count = 0;
		}
		break;

	case FM_OVER_INIT:
		/* SẴZ΂΂Zɕϊ܂B */
		memset(baras, 0, sizeof baras);
		bara = baras;
		cell = &cells[0][0];
		for(y = 0; y < FIELD_ROWS; y++) {
			for(x = 0; x < FIELD_COLS; x++) {
				if(cell->color != 0) {
					bara->color = cell->color;
					bara->x = (FIELD_X + CELL_UNIT * x) << 8;
					bara->y = (FIELD_Y + CELL_UNIT * y) << 8;
					bara->vx = RAND(field->seed) - 128;
					bara++;
					cell->color = 0;
				}
				cell++;
			}
		}
		field->mode = FM_OVER;
		field->count = 0;
		break;

	case FM_OVER:
		bara = baras;
		while(bara->color != 0) {
			bara->x += bara->vx;
			bara->y += bara->vy;
			if(bara->bounce < 1) { /* 񂾂oEX */
				if((bara->y >> 8) <= DISP_Y - CELL_UNIT) {
					bara->vy += BARA_G;
				} else {
					bara->y = (DISP_Y - CELL_UNIT) << 8;
					bara->vy = -(int)(bara->vy * BARA_R);
					bara->bounce++;
				}
			} else {
				if((bara->y >> 8) < DISP_Y) {
					bara->vy += BARA_G;
				} else {
					bara->color = 0;
				}
			}
			bara++;
		}
		/* FM_OVER͏I܂B */
		break;

	default:
		DIE();
	}

	return 1;
}

/****************************************************************************
 *	flip_check
 ****************************************************************************/

static int
flip_check(CELL cells[FIELD_ROWS][FIELD_COLS], int x, int y, FLIP flip[10])
{
	int result, i;

	result = 0;

	for(i = 0; i < 10; i++) {
		flip->x = x + flip_check_tbl[i][0];
		flip->y = y + flip_check_tbl[i][1];
		flip->vx = flip_check_tbl[i][2];
		flip->vy = flip_check_tbl[i][3];
		flip->n = flip_check_1(cells, flip->x, flip->y, flip->vx, flip->vy);
		if(flip->n) {
			flip->clear = (flip->y + flip->vy * (flip->n + 1) == FIELD_ROWS - 1);
			flip++;
			result++;
		}
	}

	return result;
}

/****************************************************************************
 *	flip_check_1
 ****************************************************************************/

static int
flip_check_1(CELL cells[FIELD_ROWS][FIELD_COLS], int x, int y, int vx, int vy)
{
	int n;
	char color;

	n = 0;
	color = cells[y][x].color;

	for(;;) {
		x += vx;
		y += vy;
		if(x < 0 || x >= FIELD_COLS) return 0;
		if(y < 0 || y >= FIELD_ROWS) return 0;
		if(cells[y][x].color == 0) return 0;
		if(cells[y][x].color == color) break;
		n++;
	}

	return n;
}

/****************************************************************************
 *	flip_check_1
 ****************************************************************************/

static void
flip_seek(FIELD* field)
{
	int x, y;
	CELL cells[FIELD_ROWS][FIELD_COLS];

	/* ̂܂ܗꍇ́ÃtB[h쐬܂B */
	memcpy(cells, field->cells, sizeof cells);
	x = field->x;
	y = field->y;
	while(y < FIELD_ROWS - 3 &&
	      cells[y + 2][x + 0].color == 0 && cells[y + 2][x + 1].color == 0) y++;
	cells[y + 0][x + 0].color = field->ctrl[0][0];
	cells[y + 0][x + 1].color = field->ctrl[0][1];
	cells[y + 1][x + 0].color = field->ctrl[1][0];
	cells[y + 1][x + 1].color = field->ctrl[1][1];
	/* flip_check()colorȂ̂ŁAmode͐ݒsv */

	/* ]i[܂B */
	field->n_flip = flip_check(cells, x, y, field->flip);
}

/****************************************************************************
 *	star_new
 ****************************************************************************/

static void
star_new(int x, int y)
{
	int i;
	STAR* star;

	star = stars;
	for(i = 0; i < STARS_MAX; i++) {
		if(!star->use) {
			star->use = 1;
			star->x = x;
			star->y = y;
			star->count = 0;
			break;
		}
		star++;
	}
}

/****************************************************************************
 *	star_draw
 ****************************************************************************/

static void
star_draw(STAR* star)
{
	PAT* pat;
	DRAW_OBJECT obj;

	pat = &star_pat[star->count];
	pceLCDSetObject(&obj, &bmp[0], star->x, star->y, pat->x, pat->y, pat->w, pat->h, DRW_NOMAL);
	pceLCDDrawObject(obj);

	if(++star->count == 6) star->use = 0;
}

